/***********************************************************
Copyright 1991 by the Massachusetts Institute of Technology,
Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

******************************************************************/

/*
 * Author:	 Ralph R. Swick
 * Date Written: May, 1991
 */


/* 
 * Pasteup.c - Pasteup composite widget
 * 
 */

#include	<X11/IntrinsicP.h>
#include	<X11/StringDefs.h>
#include	"PasteupP.h"

/****************************************************************
 *
 * Full class record constant
 *
 ****************************************************************/

static XtGeometryResult GeometryManager();
static void ChangeManaged();
static XtGeometryResult PreferredSize();

PasteupClassRec pasteupClassRec = {
  {
/* core_class fields      */
    /* superclass         */    (WidgetClass) &compositeClassRec,
    /* class_name         */    "Pasteup",
    /* widget_size        */    sizeof(PasteupRec),
    /* class_initialize   */    NULL,
    /* class_part_init    */	NULL,
    /* class_inited       */	FALSE,
    /* initialize         */    NULL,
    /* initialize_hook    */	NULL,
    /* realize            */    XtInheritRealize,
    /* actions            */    NULL,
    /* num_actions	  */	0,
    /* resources          */    NULL,
    /* num_resources      */    0,
    /* xrm_class          */    NULLQUARK,
    /* compress_motion	  */	FALSE,
    /* compress_exposure  */	FALSE,
    /* compress_enterleave*/	FALSE,
    /* visible_interest   */    FALSE,
    /* destroy            */    NULL,
    /* resize             */    NULL,
    /* expose             */    NULL,
    /* set_values         */    NULL,
    /* set_values_hook    */	NULL,
    /* set_values_almost  */    XtInheritSetValuesAlmost,
    /* get_values_hook    */	NULL,
    /* accept_focus       */    NULL,
    /* version            */	XtVersion,
    /* callback_private   */    NULL,
    /* tm_table           */    NULL,
    /* query_geometry     */	PreferredSize,
    /* display_accelerator*/	XtInheritDisplayAccelerator,
    /* extension          */	NULL
  },{
/* composite_class fields */
    /* geometry_manager   */    GeometryManager,
    /* change_managed     */    ChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension          */	NULL
  },{
/* Pasteup class fields */
    /* empty		  */	0,
  }
};

WidgetClass pasteupWidgetClass = (WidgetClass)&pasteupClassRec;


/****************************************************************
 *
 * Private Routines
 *
 ****************************************************************/

/*
 *
 * Calculate preferred size; just large enough to not clip children
 *
 */

static XtGeometryResult PreferredSize(widget, constraint, preferred)
    Widget widget;
    XtWidgetGeometry *constraint, *preferred;
{
    PasteupWidget w = (PasteupWidget)widget;
    Dimension width, height;
    int i;

    width = height = 1;

    for (i = 0; i < w->composite.num_children; i++) {
	Widget child = w->composite.children[i];
	int edge;
	if (!XtIsManaged(child)) continue;
	if (width < (edge = child->core.x + child->core.width +
			    2*child->core.border_width))
	    width = edge;

	if (height < (edge = child->core.y + child->core.height +
			     2*child->core.border_width))
	    height = edge;
    }

    preferred->request_mode = CWWidth | CWHeight;
    preferred->width = width;
    preferred->height = height;

    if ((constraint->request_mode & (CWWidth|CWHeight)) == (CWWidth|CWHeight)
	&& constraint->width == width
	&& constraint->height == height)
	return XtGeometryYes;
    else
	if (w->core.width == width && w->core.height == height)
	    return XtGeometryNo;
    else
	return XtGeometryAlmost;

}

/* Try to avoid clipping any child */

static void GrowIfNecessary(w)
    Widget w;
{
    XtWidgetGeometry constraint, preferred;

    constraint.request_mode = 0;
    (void) PreferredSize(w, &constraint, &preferred);

    if (w->core.width > preferred.width)
	preferred.width = w->core.width;

    if (w->core.height > preferred.height)
	preferred.height = w->core.height;

    if ((preferred.width != w->core.width) ||
	(preferred.height != w->core.height)) {
	if (XtMakeGeometryRequest(w, &preferred, &preferred) ==
	    XtGeometryAlmost)
	    XtMakeGeometryRequest(w, &preferred, &preferred);
    }
}

/*
 *
 * Geometry Manager
 *
 * all changes are permitted; will try to grow to avoid clipping if necessary
 *
 */

/*ARGSUSED*/
static XtGeometryResult GeometryManager(w, request, reply)
    Widget		w;
    XtWidgetGeometry	*request;
    XtWidgetGeometry	*reply;	/* RETURN; unused */

{

#define Put(flag, field) \
    if (request->request_mode & flag) w->core.field = request->field;

    if (!(request->request_mode & XtCWQueryOnly)) {
	Put(CWX, x);
	Put(CWY, y);
	Put(CWWidth, width);
	Put(CWHeight, height);
	Put(CWBorderWidth, border_width);

	GrowIfNecessary(XtParent(w));
    }

    return XtGeometryYes;
}

static void ChangeManaged(w)
    Widget w;
{
    GrowIfNecessary(w);
}
